import pytest
import equinox as eqx
from equinox import nn
from jaxtyping import PRNGKeyArray
from jaxamp import (
    amp,
    DynamicScalerState,
    dynamic_scale_grad,
    dynamic_scale_value_and_grad,
    amp_stop,
    default_amp_policy,
    use_original_precision,
)
import jax
from jax import numpy as jnp
from jax.random import PRNGKey


class HalfLinear(eqx.Module):
    l1: nn.Linear

    def __init__(self, in_features, out_features, *, key):
        self.l1 = nn.Linear(in_features, out_features, use_bias=True, key=key)

    def __call__(self, x, *, key=None):
        l1 = eqx.tree_at(
            lambda tree: (tree.weight, tree.bias),
            self.l1,
            replace_fn=lambda x: x.astype(jnp.float16),
        )
        x = x.astype(jnp.float16)
        y = l1(x)
        return y.astype(jnp.float32)


class StopBiasLinear(eqx.Module):
    l1: nn.Linear

    def __init__(self, in_features, out_features, *, key):
        self.l1 = nn.Linear(in_features, out_features, use_bias=True, key=key)

    def __call__(self, x, *, key=None):
        y = self.l1.weight @ x
        with amp_stop():
            y = y + self.l1.bias
        return y


class HalfWeightLinear(eqx.Module):
    l1: nn.Linear

    def __init__(self, in_features, out_features, *, key):
        self.l1 = nn.Linear(in_features, out_features, use_bias=True, key=key)
        # print("l1: ", self.l1)

    def __call__(self, x, *, key=None):
        l1 = eqx.tree_at(
            lambda tree: (tree.weight),
            self.l1,
            replace_fn=lambda x: x.astype(jnp.float16),
        )
        # print("self.l1 weight: ", self.l1.weight)
        # print("l1 weight: ", l1.weight)
        x = x.astype(jnp.float16)
        y = l1(x)
        return y.astype(jnp.float32)


def test_linear_amp():
    key = PRNGKey(0)
    half = HalfLinear(2, 10, key=key)  # HalfNet(key)
    full = nn.Linear(2, 10, use_bias=True, key=key)
    bias_stopped = StopBiasLinear(2, 10, key=key)
    half_weight = HalfWeightLinear(2, 10, key=key)
    # full = FullNet(key)
    amp_full = amp(full)
    amp_half = amp(half)
    amp_stopped = amp(amp_stop(full))
    amp_bias_stopped = amp(bias_stopped)
    amp_no_linear = amp(
        full, amp_policy=default_amp_policy | {"eqx.nn.Linear": use_original_precision}
    )

    x = jnp.array([5.0, 2.0])

    amp_stopped_out = amp_stopped(x)
    amp_bias_out = amp_bias_stopped(x)
    half_weight_out = half_weight(x)
    bias_out = bias_stopped(x)
    half_out = half(x)
    amp_full_out = amp_full(x)
    full_out = full(x)
    amp_half_out = amp_half(x)
    amp_no_linear_out = amp_no_linear(x)

    assert jnp.allclose(half_out, amp_full_out)
    assert jnp.allclose(half_out, amp_half_out)
    assert jnp.allclose(half_weight_out, amp_bias_out)
    assert jnp.allclose(bias_out, full_out)
    assert not jnp.allclose(full_out, half_out)
    assert jnp.allclose(full_out, amp_stopped_out)
    assert jnp.allclose(full_out, amp_no_linear_out)


def test_static():
    key = PRNGKey(0)
    bn, state = eqx.nn.make_with_state(eqx.nn.BatchNorm)(
        2, axis_name="batch"
    )  # , key=key)

    def get_batch_value(bn, state, data):
        ans, state = bn(data, state)
        return ans, state

    def get_value(bn, state, data):
        vmap_value = jax.vmap(
            get_batch_value,
            axis_name="batch",
            in_axes=(None, None, 0),
            out_axes=(0, None),
        )
        return vmap_value(bn, state, data)

    amp_get_value = amp(get_value)

    data = jnp.ones((3, 2, 4))
    value, _ = get_value(bn, state, data)
    amp_value, _ = amp_get_value(bn, state, data)

    # with all ones there will be no floating point error
    assert jnp.allclose(value, amp_value)


@pytest.mark.parametrize("tx", ["grad", "value_and_grad"])
@pytest.mark.parametrize("f", [True, False])
@pytest.mark.parametrize("aux", [True, False])
def test_dynamic_scaler(tx, f, aux):
    if tx == "grad":
        transform = dynamic_scale_grad
    elif tx == "value_and_grad":
        transform = dynamic_scale_value_and_grad

    def func(x):
        if aux:
            return x**3, 1.0
        return x**3

    x = jnp.array(2.0, dtype=jnp.float16)

    scaler_state = DynamicScalerState(scaler=2**15, patience=10)
    grad_fn = jax.jit(transform(func, filter=f, has_aux=aux, redo_on_nan=100))
    scaler_state, g = grad_fn(x, dynamic_scaler_state=scaler_state)
    if tx == "value_and_grad" and aux:
        (v, a), g = g
    if tx == "value_and_grad" and not aux:
        v, g = g
    if tx == "grad" and aux:
        g, a = g
    if aux:
        assert jnp.allclose(a, 1.0)
    if tx == "value_and_grad":
        assert jnp.allclose(v, 8.0)
    assert jnp.allclose(g, 12.0)
    assert jnp.allclose(scaler_state.scaler, 2**12)
    assert jnp.allclose(scaler_state.count, 1.0)

    for i in range(9):
        scaler_state, g = grad_fn(x, dynamic_scaler_state=scaler_state)
        assert jnp.allclose(scaler_state.scaler, 2**12)
        assert jnp.allclose(scaler_state.count, i + 2.0)
        if tx == "value_and_grad" and aux:
            (v, a), g = g
        if tx == "value_and_grad" and not aux:
            v, g = g
        if tx == "grad" and aux:
            g, a = g

        if aux:
            assert jnp.allclose(a, 1.0)
        if tx == "value_and_grad":
            assert jnp.allclose(v, 8.0)

        assert jnp.allclose(g, 12.0)

    scaler_state, g = grad_fn(x, dynamic_scaler_state=scaler_state)
    assert jnp.allclose(scaler_state.scaler, 2**13)
    assert jnp.allclose(scaler_state.count, 0.0)

    scaler_state, g = grad_fn(x, dynamic_scaler_state=scaler_state)
    assert jnp.allclose(scaler_state.scaler, 2**12)
    assert jnp.allclose(scaler_state.count, 1.0)

    scaler_state = DynamicScalerState(scaler=2**15)
    grad_fn = jax.jit(transform(func, filter=f, has_aux=aux, redo_on_nan=0))
    scaler_state, g = grad_fn(x, dynamic_scaler_state=scaler_state)
    assert jnp.allclose(scaler_state.scaler, 2**14)
    assert jnp.allclose(scaler_state.count, 0.0)

    scaler_state, g = grad_fn(x, dynamic_scaler_state=scaler_state)
    assert jnp.allclose(scaler_state.scaler, 2**13)
    assert jnp.allclose(scaler_state.count, 0.0)

    scaler_state, g = grad_fn(x, dynamic_scaler_state=scaler_state)
    assert jnp.allclose(scaler_state.scaler, 2**12)
    assert jnp.allclose(scaler_state.count, 0.0)

    scaler_state, g = grad_fn(x, dynamic_scaler_state=scaler_state)
    assert jnp.allclose(scaler_state.scaler, 2**12)
    assert jnp.allclose(scaler_state.count, 1.0)

    if tx == "value_and_grad" and aux:
        (v, a), g = g
    if tx == "value_and_grad" and not aux:
        v, g = g
    if tx == "grad" and aux:
        g, a = g

    if aux:
        assert jnp.allclose(a, 1.0)
    if tx == "value_and_grad":
        assert jnp.allclose(v, 8.0)

    assert jnp.allclose(g, 12.0)



# SPDX-FileCopyrightText: 2023-present Ashok Cutkosky <ashok@cutkosky.com>
#
# SPDX-License-Identifier: Apache-2.0

from ._dynamic_scalar import (
    DynamicScalerState,
    dynamic_scale_tx,
    dynamic_scale_value_and_grad,
    dynamic_scale_grad,
)
from ._amp import (
    amp,
    amp_stop,
    default_amp_policy,
    use_original_precision,
    use_compute_precision,
)



import jax
import jax.numpy as jnp
from functools import wraps

from jax import lax
from jax._src.util import (
    safe_map,
    curry,
)
from jax._src.core import Jaxpr, Atom, Literal, Var, last_used, clean_up_dead_vars
from jax._src import source_info_util
from typing import (
    Any,
    Callable,
    ClassVar,
    DefaultDict,
    Generic,
    NamedTuple,
    TypeVar,
    Union,
    Sequence,
    Type,
    Dict,
    Optional,
    cast,
    overload,
)
from types import MappingProxyType
from collections import defaultdict
import ml_dtypes
import equinox as eqx
from jaxtyping import PyTree
from jax import tree_util as jtu

map, unsafe_map = safe_map, map

precision_ordering = [
    jnp.float64,
    jnp.float32,
    ml_dtypes.bfloat16,
    jnp.float16,
]


def find_widest_dtype(dtype_list: Sequence[Type]) -> Type:
    for dtype in precision_ordering:
        if dtype in dtype_list:
            return dtype

    return dtype_list[0]


def use_original_precision(
    compute_dtype: Type, original_dtypes: Sequence[Type], *invars: Sequence[Any], **bind_params: Dict[str, Any]
) -> (Sequence[Any], Dict[str, Any]):
    return [cast_if_float(dtype, inv) for inv, dtype in zip(invars, original_dtypes)], bind_params


def use_widest_precision(
    compute_dtype: Type, original_dtypes: Sequence[Type], *invars: Sequence[Any], **bind_params: Dict[str, Any]
) -> (Sequence[Any], Dict[str, Any]):
    dtype = find_widest_dtype(original_dtypes)
    return use_compute_precision(dtype, original_dtypes, *invars, **bind_params)


def cast_if_float(dtype: Type, value: Any) -> Any:
    if eqx.is_inexact_array(value):
        if value.dtype != dtype:
            return value.astype(dtype)
    return value


def cast_tree(dtype: Type, tree: PyTree) -> PyTree:
    return jtu.tree_map(lambda v: cast_if_float(dtype, v), tree)


def use_precision(override_dtype: Type) -> Callable:
    def overridden_precision(compute_dtype, *args, **kwargs):
        return use_compute_precision(override_dtype, *args, **kwargs)

    return overridden_precision


def use_compute_precision(
    compute_dtype: Type, original_dtypes: Sequence[Type], *invars: Sequence[Any], **bind_params: Dict[str, Any]
) -> (Sequence[Any], Dict[str, Any]):
    for invar in invars:
        # if anything is an integer, let's not cast anything.
        if not eqx.is_inexact_array(invar):
            return invars, bind_params
    invars = cast_tree(compute_dtype, invars)
    bind_params = cast_tree(compute_dtype, bind_params)
    bind_params = dict(bind_params)
    if "preferred_element_type" in bind_params:
        bind_params["preferred_element_type"] = compute_dtype
    return invars, bind_params


low_precision_primitives = [
    lax.dot_general_p,
    lax.add_p,
    lax.sub_p,
    lax.conv_general_dilated_p,
    'eqx.nn.Linear',
    'eqx.nn.Conv2d',
    'eqx.nn.Conv',
]

high_precision_primitives = [
    lax.mul_p,
    lax.div_p,
    'eqx.nn.BatchNorm',
    'eqx.nn.LayerNorm',
    'eqx.nn.SpectralNorm',
]

# this dict specifies how different ops should be treated
# for AMP. Right now it is very simple: we move to low
# precision for additions, subtractions and
# tensor contractions  (e.g. matrix multiplies, convs etc).
# Otherwise we do not do anything.
default_amp_policy = (
    {op: use_compute_precision for op in low_precision_primitives}
    | {
        "amp_default": use_original_precision,
        "amp_stop": use_original_precision,
    }
    | {op: use_precision(jnp.float32) for op in high_precision_primitives}
)


def amp_stop(f: Callable = None):
    context = jax.named_scope("amp_stop")
    if f is None:
        return context
    else:
        return context(f)


def amp_eval_jaxpr(
    compute_dtype: Type, amp_policy: Dict[Any, Any], closed_jaxpr: Jaxpr, *args, propagate_source_info=True
) -> Any:
    jaxpr = closed_jaxpr.jaxpr

    def read(v: Atom) -> Any:
        return v.val if isinstance(v, Literal) else env[v]

    def write(v: Var, val: Any) -> None:
        env[v] = val

    env: dict[Var, Any] = {}
    map(write, jaxpr.constvars, closed_jaxpr.consts)
    map(write, jaxpr.invars, args)
    lu = last_used(jaxpr)
    for eqn in jaxpr.eqns:
        subfuns, bind_params = eqn.primitive.get_bind_params(eqn.params)
        name_stack = source_info_util.current_name_stack() + eqn.source_info.name_stack
        traceback = eqn.source_info.traceback if propagate_source_info else None
        with source_info_util.user_context(traceback, name_stack=name_stack):
            invars = map(read, eqn.invars)
            scopes = str(name_stack).split("/")
            scopes.append(eqn.primitive)
            scopes.append("amp_default")
            for scope in scopes:
                if scope in amp_policy:
                    invar_dtypes = map(lambda x: x.aval.dtype, eqn.invars)
                    invars, bind_params = amp_policy[scope](compute_dtype, invar_dtypes, *invars, **bind_params)
                    break
            outvar_dtypes = map(lambda x: x.aval.dtype, eqn.outvars)
            ans = eqn.primitive.bind(*subfuns, *invars, **bind_params)
            if not eqn.primitive.multiple_results:
                ans = [ans]
        map(write, eqn.outvars, ans)
        clean_up_dead_vars(eqn, env, lu)
    ans = map(read, jaxpr.outvars)
    ans_dtypes = map(lambda x: x.aval.dtype, jaxpr.outvars)
    ans = map(lambda x, t: x.astype(t), ans, ans_dtypes)
    return ans


def amp(
    sentinal: Optional[Callable] = None,
    *,
    compute_dtype: Type = jnp.float16,
    amp_policy: Dict = default_amp_policy,
) -> Callable:
    if sentinal is not None:
        return amp(compute_dtype=compute_dtype, amp_policy=amp_policy)(
            sentinal
        )

    def decorator(fn: Callable) -> Any:
        @wraps(fn)
        def wrapped_fn(*args, **kwargs):
            flat_args, flat_args_treedef = jtu.tree_flatten((args, kwargs))

            array_idx = [i for i, x in enumerate(flat_args) if eqx.is_array(x) or eqx.is_inexact_array_like(x)]
            static_idx = [i for i, x in enumerate(flat_args) if not (eqx.is_array(x) or eqx.is_inexact_array_like(x))]

            array_args = [flat_args[i] for i in array_idx]
            static_args = [flat_args[i] for i in static_idx]

            def flat_fn(*flat_args):
                args, kwargs = jtu.tree_unflatten(flat_args_treedef, flat_args)
                return fn(*args, **kwargs)

            jaxpr_generator = jax.make_jaxpr(flat_fn, static_argnums=static_idx, return_shape=True)

            closed_jaxpr, output_shape = jaxpr_generator(*flat_args)
            _, output_treedef = jtu.tree_flatten(output_shape)
            flat_out = amp_eval_jaxpr(compute_dtype, amp_policy, closed_jaxpr, *array_args)
            ans = jtu.tree_unflatten(output_treedef, flat_out)

            # For some reason, not including the following commented-out line
            # (which should be a no-op since the output is never accessed) causes jit to
            # recompile a few more times than it should really need to. Outside of a jit
            # context though, this line is very expensive. So, we'll leave it out
            # and just eat the small number of extra compiles.
            # It would be great to understand why those compiles are happening though...

            # nonjaxpr_ans = flat_fn(*flat_args)

            return ans

        return wrapped_fn

    return decorator



import equinox as eqx
from equinox import nn
from typing import Callable, Type, Tuple, Any, Optional, List, NamedTuple
import ml_dtypes
from jax import random as jrandom
from jax import numpy as jnp
import jax
from jax import tree_util as jtu
from jaxtyping import Array, PyTree
from dataclasses import dataclass, astuple, asdict
from functools import wraps
from jax import core
from jax import lax
from jax._src.util import safe_map


def all_finite(tree: PyTree) -> jax.Array:
    # from JMP
    leaves = jtu.tree_leaves(tree)
    if not leaves:
        return jnp.array(True)
    finite = [jnp.all(jnp.isfinite(x)) for x in leaves]
    return jnp.stack(list(finite)).all()


class DynamicScalerState(NamedTuple):
    patience: jax.Array = jnp.array(
        2000
    )  # number of non-inf/NaN iterations to wait before increasing the scaler
    adjust_factor: jax.Array = jnp.array(
        2.0
    )  # When increasing or decreasing the scaler, multiply or divide by this factor.
    scaler: jax.Array = jnp.array(2**15, dtype=jnp.float32)  # current scaler value
    count: jax.Array = jnp.array(
        0
    )  # number of non-inf/NaN iterations since the scaler was last increased.


def increment_state(state: DynamicScalerState) -> DynamicScalerState:
    new_state = jax.lax.cond(
        state.count >= state.patience,
        lambda state: DynamicScalerState(
            state.patience, state.adjust_factor, state.scaler * state.adjust_factor, 0.0
        ),
        lambda state: DynamicScalerState(
            state.patience, state.adjust_factor, state.scaler, state.count + 1.0
        ),
        state,
    )

    return new_state


def decrease_scaler(state: DynamicScalerState) -> DynamicScalerState:
    return DynamicScalerState(
        state.patience, state.adjust_factor, state.scaler / state.adjust_factor, 0.0
    )


def default_unscale_fn(results: Any, state: DynamicScalerState) -> Any:
    return jtu.tree_map(lambda x: x.astype(state.scaler.dtype) / state.scaler, results)


def value_and_grad_aux_unscale_fn(results: Any, state: DynamicScalerState) -> Any:
    (value, aux), grad = results
    value = default_unscale_fn(value, state)
    grad = default_unscale_fn(grad, state)
    return (value, aux), grad


def grad_aux_unscale_fn(results: Any, state: DynamicScalerState) -> Any:
    grad, aux = results
    grad = default_unscale_fn(grad, state)
    return grad, aux


def default_scale_fn(result: Any, state: DynamicScalerState) -> Any:
    if not eqx.is_array_like(result):
        value, aux = result
        value = state.scaler.astype(value.dtype) * value
        result = (value, aux)
    else:
        result = result * state.scaler.astype(result.dtype)
    return result


def dynamic_scale_tx(
    transform: Callable[Any, Any],
    redo_on_nan: int = 0,
    unscale_fn: Callable = default_unscale_fn,
    scale_fn: Callable = default_scale_fn,
):
    def scaled_transform(fun, *args, **kwargs):
        @wraps(fun)
        def scaled_fun(*f_args, _dynamic_scaler_state: DynamicScalerState, **f_kwargs):
            result = fun(*f_args, **f_kwargs)
            return scale_fn(result, _dynamic_scaler_state)

        transformed_fn = transform(scaled_fun, *args, **kwargs)

        def maybe_adjust_scaler(
            *f_args, dynamic_scaler_state: DynamicScalerState, **f_kwargs
        ):
            # avoid type mismatch complaints later in case state is initialized
            # with raw python ints/floats
            state = jtu.tree_map(
                lambda x: jnp.array(x).astype(jnp.float32), dynamic_scaler_state
            )

            results = transformed_fn(*f_args, _dynamic_scaler_state=state, **f_kwargs)

            results = unscale_fn(results, state)

            new_state = jax.lax.cond(
                all_finite(results), increment_state, decrease_scaler, state
            )

            return new_state, results

        if redo_on_nan == 0:
            return maybe_adjust_scaler

        def adjust_scaler_until_finite(
            *f_args, dynamic_scaler_state: DynamicScalerState, **f_kwargs
        ):
            redo_count = jnp.array(0, jnp.int32)

            new_state, results = maybe_adjust_scaler(
                *f_args, dynamic_scaler_state=dynamic_scaler_state, **f_kwargs
            )

            init_val = (new_state, results, redo_count)

            def cond_fun(state__results__redo_count):
                state, results, redo_count = state__results__redo_count
                return jnp.logical_and(
                    jnp.logical_not(all_finite(results)), redo_count < redo_on_nan
                )

            def body_fun(state__results__redo_count):
                state, results, redo_count = state__results__redo_count
                state, results = maybe_adjust_scaler(
                    *f_args, dynamic_scaler_state=state, **f_kwargs
                )
                return (state, results, redo_count + 1)

            new_state, results, redo_count = jax.lax.while_loop(
                cond_fun=cond_fun,
                body_fun=body_fun,
                init_val=init_val,
            )
            return new_state, results

        return adjust_scaler_until_finite

    return scaled_transform


def dynamic_scale_grad(
    fun: Callable,
    *,
    has_aux: bool = False,
    redo_on_nan: int = 0,
    filter=True,
    **kwargs
):
    '''
    apply dynamic scalar to the grad function.

    Args:
        fun: function to differentiate
        has_aux: same meaning as in jax.grad
        redo_on_nan: if the output is nan, we will decrease the scaler
            and recompute this many times. If the output remains nan, give up
            and return it.
        filter: if True, differentiate with equinox.filter_grad, otherwise use jax.grad

    Returns:
        grad_fn: a function that behaves like the output of jax.grad except:
            1. has an extra required keyword argument dynamic_scaler_state
            2. the return value is now a tuple (next_dynamic_scaler_state, grads)
                of (next_dynamic_scaler_state, (grads, aux)) if has_aux=True
    '''
    if has_aux:
        unscale_fn = grad_aux_unscale_fn
    else:
        unscale_fn = default_unscale_fn

    if filter:
        tx = eqx.filter_grad
    else:
        tx = jax.grad

    grad_fn = dynamic_scale_tx(tx, redo_on_nan=redo_on_nan, unscale_fn=unscale_fn)(
        fun, has_aux=has_aux, **kwargs
    )
    return grad_fn


def dynamic_scale_value_and_grad(
    fun: Callable,
    *,
    has_aux: bool = False,
    redo_on_nan: bool = 10,
    filter=True,
    **kwargs
):
    '''
    apply dynamic scalar to the value_and_grad function.

    Args:
        fun: function to differentiate
        has_aux: same meaning as in jax.grad
        redo_on_nan: if the output is nan, we will decrease the scaler
            and recompute this many times. If the output remains nan, give up
            and return it.
        filter: if True, differentiate with equinox.filter_value_and_grad, otherwise use jax.value_and_grad

    Returns:
        grad_fn: a function that behaves like the output of jax.value_and_grad except:
            1. has an extra required keyword argument dynamic_scaler_state
            2. the return value is now a tuple (next_dynamic_scaler_state, (value, grads))
                of (next_dynamic_scaler_state, ((value, aux), grads)) if has_aux=True
    '''
    if has_aux:
        unscale_fn = value_and_grad_aux_unscale_fn
    else:
        unscale_fn = default_unscale_fn
    if filter:
        tx = eqx.filter_value_and_grad
    else:
        tx = jax.value_and_grad

    grad_fn = dynamic_scale_tx(tx, redo_on_nan=redo_on_nan, unscale_fn=unscale_fn)(
        fun, has_aux=has_aux, **kwargs
    )
    return grad_fn




import jax
from jax import nn
from functools import wraps
from dataclasses import dataclass


MixedPrecisionContext

def __getattr__(name):
    nn_attr = getattr(nn, name)

    if not callable(nn_attr):
        return nn_attr

    @wraps(nn_attr)
    def wrapped_fn
        
    def wrap(*args, **kwargs):
